-- Streaming Outliers Clustering
-- Implements the algorithm from the paper, except for the initialization of r
-- (r has to be specified), without parallelization.
module SOC where
import Metric
import OC
import Data.List ((\\))

-- A stored center with support.
data SCtr pt = SCtr {
	scCtr :: pt,
	scSupport :: [pt]
} deriving Show

-- Do centers sc1 and sc2 conflict based on the r value in soc?
conflicts soc sc1 sc2 = any (\s1 -> any (\s2 -> dist s1 s2 <= 2 * socAlphaR soc)
	(scSupport sc2)) (scSupport sc1)

-- The state of a streaming outliers clusterer.
data Metric pt => SOC pt = SOC {
	socK :: Int,
	socZ :: Int,
	socR :: Double,
	socSCs :: [SCtr pt], -- Stored centers
	socFPs :: [pt], -- Free points
	socOCs :: [pt] -- Centers chosen by the last run of the offline algorithm.
} deriving Show {- TEMP -}

-- Multiples of the current r.
socAlphaR soc = 4 * socR soc
socBetaR soc = 8 * socR soc
socEtaR soc = 16 * socR soc

-- Current number of stored cluster centers
socEll soc = length (socSCs soc)

-- Start a clustering run with the given k, z, and r0.
socStart k z r0 = SOC {socK = k, socZ = z, socR=r0, socSCs=[], socFPs=[], socOCs=[]}

-- Get the clustering for the input provided so far.
socFinish soc = (socEtaR soc, map scCtr (socSCs soc) ++ socOCs soc)

-- Add a list of points to the clustering run.
socAdd soc newPts =
	-- Add the new points as free points and perform the procedure from Step 1.
	socStep1 soc{socFPs = (socFPs soc ++ newPts)}

-- Step 1 of the algorithm.
socStep1 soc = socStep2 soc{socFPs =
	-- Keep only free points that are at least eta r from every stored center.
	filter (\p -> all (\c -> dist (scCtr c) p > socEtaR soc)
		(socSCs soc))
	(socFPs soc)}

-- Step 2 of the algorithm.
socStep2 soc =
	-- Identify potential centers p that have >= z+1 free points nearby.
	let nearFPs p = filter (\p2 -> dist p p2 <= socBetaR soc) (socFPs soc) in
	case filter (\p -> length (nearFPs p) >= socZ soc + 1) (socFPs soc) of
		[] -> socStep3 soc
		newSC : _ ->
			-- Choose z+1 support points.
			let support = take (socZ soc + 1) (nearFPs newSC) in
			-- Add the new center with support.
			-- The support points will automatically be removed from
			-- the list of free points in socStep1.
			socStep1 soc{socSCs = SCtr newSC support : socSCs soc}

-- Step 3 of the algorithm.
socStep3 soc =
	-- Two checks and offline algorithm...
	let newOCs = tryOutliersCluster (socAlphaR soc) (socK soc - socEll soc) (socFPs soc) in
	if socEll soc <= socK soc
		&& length (socFPs soc) <= (socK soc - socEll soc + 1) * socZ soc
		&& length (socFPs soc) - nPtsCovered newOCs (socFPs soc) (socEtaR soc) <= socZ soc
	then -- Success; save the results of the offline algorithm.
		soc{socOCs = newOCs} {- finished -}
	else -- Raise r and reprocess the stored centers.
		socStep4 soc{socR = socAlphaR soc, socSCs = []} (socSCs soc)

-- Recursive helper for Step 4 of the algorithm.
-- Roll the centers in oldCs into soc, dropping any that conflict with centers already in soc.
socStep4 soc oldCs = case oldCs of
	[] -> -- Done reprocessing centers; go back to Step 1.
		socStep1 soc
	oc : olderCs -> -- Act on the center oc.
		if any (\ec -> conflicts soc ec oc) (socSCs soc)
		then -- Skip oc due to conflict.
			socStep4 soc olderCs
		else -- Add oc.
			socStep4 soc{socSCs = oc : socSCs soc} olderCs
